-- debug_status = 1
debug_mod_name = "CargoRocket"
debug_file = debug_mod_name .. "-debug.txt"
require("utils")
require("config")

local type_departure = 1
local type_arrival = 2

--------------------------------------------------------------------------------------
local function init_globals()
	-- initialize or update general globals of the mod
	debug_print( "init_globals " )

	global.ticks = global.ticks or 0
	global.no_chest = global.no_chest or 0
	global.force_mem = global.force_mem or {}
	global.player_mem = global.player_mem or {}
	global.transits = global.transits or {}
end

--------------------------------------------------------------------------------------
local function init_force(force)
	if global.force_mem == nil then return end
	
	-- initialize or update per force globals of the mod
	debug_print( "init_force ", force.name )
	
	global.force_mem[force.name] = global.force_mem[force.name] or {}
	local force_mem = global.force_mem[force.name]
	
	force_mem.departures = force_mem.departures or {}
	force_mem.arrivals = force_mem.arrivals or {}
end

--------------------------------------------------------------------------------------
local function init_forces()
	for _, force in pairs(game.forces) do
		init_force(force)
	end
end

--------------------------------------------------------------------------------------
local function init_player(player)
	if global.player_mem == nil then return end
	
	-- initialize or update per player globals of the mod, and reset the gui
	debug_print( "init_player ", player.name, " connected=", player.connected )
	
	global.player_mem[player.index] = global.player_mem[player.index] or {}
	
	local player_mem = global.player_mem[player.index]
	
	player_mem.opened = player_mem.opened or nil
	player_mem.chest_cur = player_mem.chest_cur or nil
	player_mem.txt_name = player_mem.txt_name or nil
	player_mem.list_no = player_mem.list_no or 1
	player_mem.list_max = player_mem.list_max or 0
end

--------------------------------------------------------------------------------------
local function init_players()
	for _, player in pairs(game.players) do
		init_player(player)
	end
end

--------------------------------------------------------------------------------------
local function on_init() 
	-- called once, the first time the mod is loaded on a game (new or existing game)
	debug_print( "on_init" )
	init_globals()
	init_forces()
	init_players()
end

script.on_init(on_init)

--------------------------------------------------------------------------------------
local function on_configuration_changed(data)
	-- detect any mod or game version change
	if data.mod_changes ~= nil then
		local changes = data.mod_changes[debug_mod_name]
		if changes ~= nil then
			debug_print( "update mod: ", debug_mod_name, " ", tostring(changes.old_version), " to ", tostring(changes.new_version) )
		
			for _,force in pairs(game.forces) do
				force.recipes["cargo-rocket-silo"].enabled = force.technologies["cargo-rockets"].researched
			end
			
			init_globals()
			init_forces()
			init_players()
			
			-- clean potentially remaining arrivals chests from 0.12
			
			for _,force in pairs(game.forces) do
				local force_mem = global.force_mem[force.name]
				for _, chest in pairs(force_mem.departures) do
					if chest.link ~= nil and not chest.link.entity.valid then
						chest.link = nil
					end
				end
				for _, chest in pairs(force_mem.arrivals) do
					if chest.entity ~= nil and not chest.entity.valid then
						if chest.mapmark ~= nil and chest.mapmark.valid then
							chest.mapmark.destroy()
						end
					end
				end
			end
		end
	end
end

script.on_configuration_changed(on_configuration_changed)

--------------------------------------------------------------------------------------

local function on_force_created(event)
	local force = event.force
	debug_print( "force created ", force.name )
	
	init_force(force)
end
	
script.on_event(defines.events.on_force_created, on_force_created)

--------------------------------------------------------------------------------------
local function on_player_created(event)
	local player = game.players[event.player_index]
	debug_print( "player created ", player.name )
	
	init_player(player)
end

script.on_event(defines.events.on_player_created, on_player_created )

--------------------------------------------------------------------------------------
local function on_player_joined_game(event)
	-- called in SP(once) and MP(every connect), eventually after on_player_created
	local player = game.players[event.player_index]
	debug_print( "player joined ", player.name )
	
	init_player(player)
end

script.on_event(defines.events.on_player_joined_game, on_player_joined_game )

--------------------------------------------------------------------------------------
local function on_creation( event )
	local ent = event.created_entity
	
	if ent.name == "cargo-departure" then
		debug_print( "creation ", ent.name )
		
		global.no_chest = global.no_chest + 1
		local pos = ent.position
		local force_mem = global.force_mem[ent.force.name]
		local chest = {
			entity = ent,
			mapmark = nil,
			type = type_departure,
			name = "#" .. global.no_chest .. " (" .. math.floor(pos.x) .. "/" .. math.floor(pos.y) .. ")",
			link = nil, -- linked arrival
			silo = nil, -- linked silo
			autolaunch = false, 
		}
		
		-- check if silo around
		local box = ent.prototype.selection_box
		local area = {{pos.x+box.left_top.x-0.8, pos.y+box.left_top.y-0.8},{pos.x+box.right_bottom.x+0.8, pos.y+box.right_bottom.y+0.8}}
		local silos = ent.surface.find_entities_filtered({area=area, type = "rocket-silo"})
		
		if silos ~= nil then
			chest.silo = silos[1]
		end

		table.insert(force_mem.departures, chest )
		debug_print( "deps=", #force_mem.departures )

	elseif ent.name == "cargo-arrival" then
		debug_print( "creation ", ent.name )
		
		global.no_chest = global.no_chest + 1
		local pos = ent.position
		local force_mem = global.force_mem[ent.force.name]
		local chest = {
			entity = ent,
			mapmark = nil,
			type = type_arrival,
			name = "#" .. global.no_chest .. " (" .. math.floor(pos.x) .. "/" .. math.floor(pos.y) .. ")",
			link = nil, -- linked departure
		}
		
		table.insert(force_mem.arrivals,chest)
		debug_print( "arrs=", #force_mem.arrivals )
		
	elseif ent.type == "rocket-silo" then
		-- debug_print( "creation ", ent.name )
		
		-- check if unlinked departure around
		local pos = ent.position
		local box = ent.prototype.selection_box
		local area = {{pos.x+box.left_top.x-0.8, pos.y+box.left_top.y-0.8},{pos.x+box.right_bottom.x+0.8, pos.y+box.right_bottom.y+0.8}}
		local chests = ent.surface.find_entities_filtered({area=area, name = "cargo-departure"})
		local force_mem = global.force_mem[ent.force.name]
		
		for _, chest in pairs(chests) do
			for _, dep in pairs(force_mem.departures) do
				if dep.entity == chest then
					dep.silo = ent
					break
				end
			end
		end
		
		-- if ent.name == "cargo-rocket-silo" then
			-- ent.operable = false
		-- end

		table.insert(force_mem.departures, dep )
	end
end

script.on_event(defines.events.on_built_entity, on_creation )
script.on_event(defines.events.on_robot_built_entity, on_creation )

--------------------------------------------------------------------------------------
local function on_destruction( event )
	local ent = event.entity
	local force_mem = global.force_mem[ent.force.name]
	
	if ent.name == "cargo-departure" then
		debug_print( "destruction ", ent.name )
		
		local force_mem = global.force_mem[ent.force.name]
		local k, chest = find_chest(force_mem.departures, ent)
		if k ~= nil then
			if chest.link ~= nil then
				chest.link.link = nil
			end
			if chest.mapmark and chest.mapmark.valid then
				chest.mapmark.destroy()
			end
			table.remove( force_mem.departures, k )
			debug_print( "deps=", #force_mem.departures )
		end
		
	elseif ent.name == "cargo-arrival" then
		debug_print( "destruction ", ent.name )
		
		local force_mem = global.force_mem[ent.force.name]
		local k, chest = find_chest(force_mem.arrivals, ent)
		-- debug_print( "k=", k )
		if k ~= nil then
			if chest.link ~= nil then
				chest.link.link = nil
			end
			if chest.mapmark and chest.mapmark.valid then
				chest.mapmark.destroy()
			end
			table.remove( force_mem.arrivals, k )
			debug_print( "arrs=", #force_mem.arrivals )
		end
		
	elseif ent.type == "rocket-silo" then
		-- debug_print( "destruction ", ent.name )
		
		local force_mem = global.force_mem[ent.force.name]
		
		for _, dep in pairs(force_mem.departures) do
			if dep.silo == ent then
				dep.silo = nil
			end
		end
	end
end

script.on_event(defines.events.on_entity_died, on_destruction )
script.on_event(defines.events.on_robot_pre_mined, on_destruction )
script.on_event(defines.events.on_preplayer_mined_item, on_destruction )

--------------------------------------------------------------------------------------

local function on_rocket_launched(event)
	local rocket = event.rocket
end

script.on_event(defines.events.on_rocket_launched, on_rocket_launched)

--------------------------------------------------------------------------------------
local function on_tick(event)
	if global.ticks % 16 == 0 then
		-- manage chests opening / closing
		
		for _, player in pairs(game.players) do
			if player.connected then
				local player_mem = global.player_mem[player.index]
				local opened = player.opened
				local force_mem, chest
				if opened and opened ~= player_mem.opened then
					if opened.name == "cargo-departure" then
						-- open departure window
						force_mem = global.force_mem[player.force.name]
						for _, chest in pairs(force_mem.departures) do
							if chest.entity == opened then
								update_gui_chest( player, player_mem, chest )
								player_mem.chest_cur = chest
								break
							end
						end
					elseif opened.name == "cargo-arrival" then
						-- open arrival window
						debug_print("open arrival")
						force_mem = global.force_mem[player.force.name]
						for _, chest in pairs(force_mem.arrivals) do
							if chest.entity == opened then
								update_gui_chest( player, player_mem, chest )
								player_mem.chest_cur = chest
								break
							end
						end
					end
				elseif opened == nil and player_mem.opened ~= nil then
					-- close chest window
					if player.gui.left.frm_cgorkt_chest ~= nil then
						player.gui.left.frm_cgorkt_chest.destroy()
					end
					if player.gui.left.frm_cgorkt_list ~= nil then
						player.gui.left.frm_cgorkt_list.destroy()
					end
					player_mem.chest_cur = nil
				end
				player_mem.opened = opened
			end
		end
		
		-- check end of transit and payload deliveries
		
		for k, transit in pairs(global.transits) do
			if game.tick > transit.tick then
				local dest = transit.destination.entity
				local force = transit.force
				
				if dest == nil or not dest.valid then
					message_force(force,{"cgorkt-gui-baddest",transit.name})
				else
					local ok = true
					for name, count in pairs(transit.contents) do
						if dest.insert({name=name,count=count}) ~= count then
							ok = false
						end
					end
					
					if ok then
						if verbose_delivery then
							message_force(force,{"cgorkt-gui-trans2",transit.name})
						end
					else
						message_force(force,{"cgorkt-gui-arrfull",transit.name})
					end
				end
				table.remove(global.transits,k)
				break
			end
		end
		
		-- check cargo rocket silo ready 
		
		-- for _, force in pairs(game.forces) do
			-- local force_mem = global.force_mem[force.name]
			
			-- for k, chest in pairs(force_mem.departures) do
				-- local silo = chest.silo
				-- if silo and silo.valid and silo.name == "carfo-rocket-silo" then
					-- if silo.rocket_parts == nb_cargo_rocket_parts then
						-- silo.operable = false
					-- end
				-- end
			-- end
		-- end
	end

	if global.ticks <= 0 then 
		global.ticks = 16 * 10
		
		-- check any departure full and possible launch
		
		for _, force in pairs(game.forces) do
			local force_mem = global.force_mem[force.name]
			
			for k, chest in pairs(force_mem.departures) do
				local ent = chest.entity
				if ent and ent.valid then
					if chest.autolaunch then
						local invent = ent.get_inventory(defines.inventory.chest)
						local bar
						local ok = true
						
						if invent.hasbar() then
							bar = invent.getbar()
						else
							bar = #invent
						end
						
						-- debug_print(chest.name, " bar ", bar )
						
						for i = 1, bar do
							local stack = invent[i]
							if stack.valid_for_read then
								local size = game.item_prototypes[stack.name].stack_size
								-- debug_print(chest.name, " ", stack.name, " ", stack.count, " ", size )
								if stack.count < size then
									ok = false
								end
							else
								ok = false
							end
						end
						
						if ok then
							-- try to launch
							launch_chest( chest )
						end
					end
				else
					table.remove(force_mem.departures,k)
				end
			end
			
			for k, chest in pairs(force_mem.arrivals) do
				local ent = chest.entity
				if ent == nil or not ent.valid then
					table.remove(force_mem.arrivals,k)
				end
			end
		end
		
		
	else
		global.ticks = global.ticks -1 
	end
end

script.on_event(defines.events.on_tick, on_tick)

--------------------------------------------------------------------------------------
local function on_gui_text_changed(event)
	local player = game.players[event.player_index]
	local event_name = event.element.name
	
	debug_print( "on_gui_text_changed ", player.name, " ", event_name )
	
	if event_name == "txt_cgorkt_nam" then -- rename chest
		local player_mem = global.player_mem[player.index]
		local chest = player_mem.chest_cur
		if chest ~= nil then
			local s = player_mem.txt_name.text
			debug_print( "on_gui_text_changed " , s )
			if s ~= "" then
				chest.name = s
				if chest.mapmark and chest.mapmark.valid then
					chest.mapmark.backer_name = s
				end
			else
				player.print( {"cgorkt-gui-noempty"} )
			end
		end
	end
end

script.on_event(defines.events.on_gui_text_changed,on_gui_text_changed)

--------------------------------------------------------------------------------------
local function on_gui_click(event)
	local player = game.players[event.player_index]
	local event_name = event.element.name
	local prefix = string.sub(event_name,1,16)
	
	-- debug_print( "player ", player.name, " click ", event_name )

	if event_name == "but_cgorkt_nam" then -- rename chest
		local player_mem = global.player_mem[player.index]
		local chest = player_mem.chest_cur
		if chest ~= nil then
			local s = player_mem.txt_name.text
			if s ~= "" then
				chest.name = s
				if chest.mapmark and chest.mapmark.valid then
					chest.mapmark.backer_name = s
				end
			else
				player.print( {"cgorkt-gui-noempty"} )
			end
		end
		
	elseif event_name == "chk_cgorkt_silo" then -- reset original checkbox state
		local player_mem = global.player_mem[player.index]
		if player_mem.chest_cur ~= nil then
			update_gui_chest(player,player_mem,player_mem.chest_cur)
		end
		
	elseif event_name == "but_cgorkt_lnk" then -- open chests list
		local player_mem = global.player_mem[player.index]
		if player_mem.chest_cur ~= nil then
			player_mem.list_no = 1
			show_list(player,player_mem,player_mem.chest_cur.type)
		end
		
	elseif event_name == "but_cgorkt_launch" then -- try to launch silo
		local player_mem = global.player_mem[player.index]
		launch_chest( player_mem.chest_cur, player )
		
	elseif event_name == "chk_cgorkt_auto" then -- switch autolaunch
		local player_mem = global.player_mem[player.index]
		if player_mem.chest_cur ~= nil then
			player_mem.chest_cur.autolaunch = not player_mem.chest_cur.autolaunch
			update_gui_chest(player,player_mem,player_mem.chest_cur)
		end
		
	elseif event_name == "chk_cgorkt_mark" then -- switch map mark
		local player_mem = global.player_mem[player.index]
		local chest = player_mem.chest_cur
		if chest ~= nil then
			local mapmark = chest.mapmark
			if mapmark ~= nil then
				if mapmark.valid then 
					mapmark.destroy()
				end
				chest.mapmark = nil
			else
				local ent = chest.entity
				if ent and ent.valid then
					mapmark = ent.surface.create_entity({name = "cargo-mapmark", force = game.forces.neutral, position = ent.position})
					mapmark.operable = false
					mapmark.active = false
					mapmark.backer_name = chest.name
					chest.mapmark = mapmark
				end
			end
			update_gui_chest(player,player_mem,player_mem.chest_cur)
		end
		
	elseif event_name == "but_cgorkt_close" then -- close chests list
		if player.gui.left.frm_cgorkt_list ~= nil then
			player.gui.left.frm_cgorkt_list.destroy()
		end
		
	elseif event_name == "but_cgorkt_pgleft" then -- chests list page left
		local player_mem = global.player_mem[player.index]
		if player_mem.chest_cur ~= nil then
			if player_mem.list_no - items_per_page >= 1 then
				player_mem.list_no = player_mem.list_no - items_per_page
				show_list(player,player_mem,player_mem.chest_cur.type)
			end
		end
		
	elseif event_name == "but_cgorkt_pgright" then -- chests list page right
		local player_mem = global.player_mem[player.index]
		if player_mem.chest_cur ~= nil then
			if player_mem.list_no + items_per_page <= player_mem.list_max then
				player_mem.list_no = player_mem.list_no + items_per_page
				show_list(player,player_mem,player_mem.chest_cur.type)
			end
		end
		
	elseif prefix == "but_cgorkt_chst_" then -- choose chest in list
		local suffix = string.sub(event_name,17)
		local player_mem = global.player_mem[player.index]
		local n_chest = tonumber(suffix)
		
		if n_chest ~= nil and player_mem.chest_cur ~= nil then
			local chest
			local force_mem = global.force_mem[player.force.name]
			if player_mem.chest_cur.type == type_departure then
				chest = force_mem.arrivals[n_chest]
			else
				chest = force_mem.departures[n_chest]
			end
			if chest ~= nil then
				player_mem.chest_cur.link = chest
				chest.link = player_mem.chest_cur
			end
		end

		update_gui_chest(player,player_mem,player_mem.chest_cur)

		if player.gui.left.frm_cgorkt_list ~= nil then
			player.gui.left.frm_cgorkt_list.destroy()
		end
	end 
end

script.on_event(defines.events.on_gui_click, on_gui_click)

--------------------------------------------------------------------------------------
local function clean_gui(gui)
	for _, guiname in pairs( gui.children_names ) do
		gui[guiname].destroy()
	end
end

--------------------------------------------------------------------------------------
function update_gui_chest(player, player_mem, chest)
	local gui1 = player.gui.left.frm_cgorkt_chest
	local s, gui2, gui3, gui4, gui5
	
	if gui1 == nil then
		gui1 = player.gui.left.add({type = "frame", name = "frm_cgorkt_chest", direction = "vertical", style = "cgorkt_frame_style"})
	end
	
	clean_gui(gui1)
	gui2 = gui1.add({type = "flow", name = "flw_cgorkt_chest", direction = "vertical", style = "cgorkt_flow_style"})

	if chest.type == type_departure then
		gui2.add({type = "label", name = "lbl_cgorkt_tit", caption = {"cgorkt-gui-dp-title"}, style = "cgorkt_label_bold_style"})
	else
		gui2.add({type = "label", name = "lbl_cgorkt_tit", caption = {"cgorkt-gui-ar-title"}, style = "cgorkt_label_bold_style"})
	end
	
	gui3 = gui2.add({type = "table", name = "tbl_cgorkt", colspan=2, style = "cgorkt_table_style"})
	
	gui3.add({type = "label", name = "lbl_cgorkt_nam", caption = {"cgorkt-gui-name"}, style = "cgorkt_label_style"})
	gui4 = gui3.add({type = "flow", name = "flw_cgorkt_nam", direction = "horizontal", style = "cgorkt_flow_style"})
	gui5 = gui4.add({type = "textfield", name = "txt_cgorkt_nam", text = chest.name, style = "cgorkt_textfield_style"})
	gui5.style.minimal_width = 150
	player_mem.txt_name = gui5
	gui4.add({type = "button", name = "but_cgorkt_nam", caption = "=", style = "cgorkt_button_style"})

	if chest.type == type_departure then
		gui3.add({type = "label", name = "lbl_cgorkt_to", caption = {"cgorkt-gui-to"}, style = "cgorkt_label_style"})
	else
		gui3.add({type = "label", name = "lbl_cgorkt_from", caption = {"cgorkt-gui-from"}, style = "cgorkt_label_style"})
	end
	if chest.link == nil then
		s = {"cgorkt-gui-undef"}
	else
		s = chest.link.name
	end
	gui3.add({type = "button", name = "but_cgorkt_lnk", caption = s, style = "cgorkt_button_style"})
	
	if chest.type == type_departure then
		gui3.add({type = "label", name = "lbl_cgorkt_silo", caption = {"cgorkt-gui-silo"}, style = "cgorkt_label_style"})
		gui3.add({type = "checkbox", name = "chk_cgorkt_silo", caption = {"cgorkt-gui-linked"}, state = (chest.silo ~= nil), style = "cgorkt_checkbox_style"})
		gui3.add({type = "label", name = "lbl_cgorkt_manu", caption = {"cgorkt-gui-manu"}, style = "cgorkt_label_style"})
		gui3.add({type = "button", name = "but_cgorkt_launch", caption = {"cgorkt-gui-launch"}, style = "cgorkt_button_style"})
		gui3.add({type = "label", name = "lbl_cgorkt_auto", caption = {"cgorkt-gui-auto"}, style = "cgorkt_label_style"})
		gui3.add({type = "checkbox", name = "chk_cgorkt_auto", state = chest.autolaunch, style = "cgorkt_checkbox_style"})
	end

	gui3.add({type = "label", name = "lbl_cgorkt_mark", caption = {"cgorkt-gui-mark"}, style = "cgorkt_label_style"})
	gui3.add({type = "checkbox", name = "chk_cgorkt_mark", state = (chest.mapmark ~= nil), style = "cgorkt_checkbox_style"})

	return(gui1)
end

--------------------------------------------------------------------------------------
function show_list(player, player_mem, typ)
	local gui1 = player.gui.left.frm_cgorkt_list
	local s, gui2, gui3, gui4, guinb
	local force_mem = global.force_mem[player.force.name]
	local chests
	
	if gui1 == nil then
		gui1 = player.gui.left.add({type = "frame", name = "frm_cgorkt_list", direction = "vertical", style = "cgorkt_frame_style"})
	end
	
	clean_gui(gui1)
	gui2 = gui1.add({type = "flow", name = "flw_cgorkt_list", direction = "vertical", style = "cgorkt_flow_style"})

	if typ == type_departure then
		gui2.add({type = "label", name = "lbl_cgorkt_tit", caption = {"cgorkt-gui-ars-title"}, style = "cgorkt_label_bold_style"})
		chests = force_mem.arrivals
	else
		gui2.add({type = "label", name = "lbl_cgorkt_tit", caption = {"cgorkt-gui-dps-title"}, style = "cgorkt_label_bold_style"})
		chests = force_mem.departures
	end

	gui3 = gui2.add({type = "flow", name = "flw_cgorkt_list1", direction = "horizontal", style = "cgorkt_flow_style"})
	gui3.add({type = "button", name = "but_cgorkt_pgleft", caption = "<<", style = "cgorkt_button_style"})
	guinb = gui3.add({type = "button", name = "but_cgorkt_pgcur", caption = " ", style = "cgorkt_button_style"})
	gui3.add({type = "button", name = "but_cgorkt_pgright", caption = ">>", style = "cgorkt_button_style"})

	gui3 = gui2.add({type = "frame", name = "frm_cgorkt_list2", direction = "vertical", style = "cgorkt_frame_style"})
	gui3 = gui3.add({type = "flow", name = "flw_cgorkt_list2", direction = "vertical", style = "cgorkt_flow_style"})

	local n = 0
	
	for k, chest in pairs(chests) do
		n = n + 1
		if k >= player_mem.list_no and k < player_mem.list_no + items_per_page then
			gui3.add({type = "button", name = "but_cgorkt_chst_" .. k, caption = chest.name, style = "cgorkt_button_thin_style"})
		end
	end

	player_mem.list_max = n
	
	guinb.caption = (1 + math.floor((player_mem.list_no-1)/items_per_page)) .. "/" .. (1 + math.floor((player_mem.list_max-1)/items_per_page))
	
	gui2.add({type = "button", name = "but_cgorkt_close", caption = {"cgorkt-gui-close"}, style = "cgorkt_button_style"})
	
	return(gui1)
end

--------------------------------------------------------------------------------------
function launch_chest( chest, player )
	-- debug_print( "try launch" )
	
	if chest == nil then 
		if player then player.print( {"cgorkt-gui-nodep"} ) end
		return(false)
	end
	
	local orig = chest.entity
	local silo = chest.silo
	local link = chest.link
	
	if orig == nil or not orig.valid then
		if player then player.print( {"cgorkt-gui-nodep"} ) end
		return(false)
	end
	
	if silo == nil or not silo.valid then
		if player then player.print( {"cgorkt-gui-nosilo"} ) end
		return(false)
	end
	
	if link == nil then
		if player then player.print( {"cgorkt-gui-nolink"} ) end
		return(false)
	end
	
	local dest = link.entity
	
	if dest == nil or not dest.valid then
		if player then player.print( {"cgorkt-gui-nolink"} ) end
		return(false)
	end
	
	local is_vanilla = (silo.name == "rocket-silo")
	-- is_vanilla = true
	
	debug_print( "vanilla=", is_vanilla )
	
	if is_vanilla and not silo.can_insert("satellite") then
		if player then player.print( {"cgorkt-gui-notready"} ) end
		return(false)
	end
	
	if is_vanilla then
		silo.insert({name = "satellite", count = 1})
	end
	
	if not silo.launch_rocket() then
		if is_vanilla then
			silo.remove_item({name = "satellite", count = 1})
		end
		if player then player.print( {"cgorkt-gui-notready"} ) end
		return(false)
	end
	
	-- silo.operable = true
	
	local invent = orig.get_inventory(defines.inventory.chest)
	local cont = invent.get_contents()
	local dist = distance(orig.position,link.entity.position)
	debug_print( "dist=", dist )
	debug_print( "cont=", #cont )
	local transit =	{
		name = chest.name .. " -> " .. link.name,
		force = orig.force,
		destination = link,
		delay = math.floor(delay_launch + dist * delay_distance),
		contents = cont,
	}
	transit.tick = game.tick + 60 * transit.delay,
	table.insert(global.transits,transit)
	invent.clear()
	if verbose_delivery then
		message_force(orig.force,{"cgorkt-gui-trans1",transit.name, transit.delay})
	end
	
	return(true)
end

--------------------------------------------------------------------------------------
function find_chest( list, ent )
	for k, chest in pairs(list) do
		if chest.entity == ent then
			return k,chest
		end
	end
	
	return nil,nil
end

--------------------------------------------------------------------------------------
function debug_guis( guip, indent )
	debug_print( indent .. string.rep("....",indent) .. " " .. guip.name )
	indent = indent+1
	for k, gui in pairs(guip.children_names) do
		debug_guis( guip[gui], indent )
	end
end