-- debug_status = 1
debug_mod_name = "CopySettingsTool"
debug_file = debug_mod_name .. "-debug.txt"
require("utils")

local wipes = { all = 1, requests = 2, filters = 3, hashbar = 4, recipes = 5, cancel = 6,
	{name = "all"},
	{name = "requests"},
	{name = "filters"},
	{name = "hashbar"},
	{name = "recipes"},
	{name = "cancel"},
}

--------------------------------------------------------------------------------------
local function init_globals()
	-- initialize or update general globals of the mod
	debug_print( "init_globals" )
	
	global.player_mem = global.player_mem or {}
end

--------------------------------------------------------------------------------------
local function init_player(player)
	if global.player_mem == nil then return end
	
	-- initialize or update per player globals of the mod, and reset the gui
	debug_print( "init_player ", player.name, " connected=", player.connected )
	
	global.player_mem[player.index] = global.player_mem[player.index] or {}
	
	local player_mem = global.player_mem[player.index]
	player_mem.player = player_mem.player or player
	player_mem.clipboard = player_mem.clipboard or nil
	player_mem.targets = player_mem.targets or nil
end

--------------------------------------------------------------------------------------
local function init_players()
	for _, player in pairs(game.players) do
		init_player(player)
	end
end

--------------------------------------------------------------------------------------
local function on_init() 
	-- called once, the first time the mod is loaded on a game (new or existing game)
	debug_print( "on_init" )
	init_globals()
	init_players()
end

script.on_init(on_init)

--------------------------------------------------------------------------------------
local function on_configuration_changed(data)
	-- detect any mod or game version change
	if data.mod_changes ~= nil then
		local changes = data.mod_changes[debug_mod_name]
		if changes ~= nil then
			debug_print( "update mod: ", debug_mod_name, " ", tostring(changes.old_version), " to ", tostring(changes.new_version) )
		
			init_globals()
			init_players()
			
			for _, force in pairs(game.forces) do
				-- force.reset_technologies()
				force.recipes["copy-settings-tool"].enabled = force.technologies["automated-construction"].researched
			end
			
			if changes.old_version and older_version(changes.old_version, "1.0.0") then
				-- migrations
			end
		end
	end
end

script.on_configuration_changed(on_configuration_changed)

--------------------------------------------------------------------------------------
local function on_player_created(event)
	-- called at player creation
	local player = game.players[event.player_index]
	debug_print( "player created ", player.name )
	
	init_player(player)
end

script.on_event(defines.events.on_player_created, on_player_created )

--------------------------------------------------------------------------------------
local function on_player_joined_game(event)
	-- called in SP(once) and MP(at every connect), eventually after on_player_created
	local player = game.players[event.player_index]
	debug_print( "player joined ", player.name )
	
	init_player(player)
end

script.on_event(defines.events.on_player_joined_game, on_player_joined_game )

--------------------------------------------------------------------------------------
local function on_player_left_game(event)
	local player = game.players[event.player_index]
	debug_print( "player left ", player.name )
	
	init_player(player)
end

script.on_event(defines.events.on_player_left_game, on_player_left_game )

--------------------------------------------------------------------------------------
function on_player_alt_selected_area(event)	
	-- copy a group of reference entities
	
	if event.item == "copy-settings-tool" then
		local player = game.players[event.player_index]
		local player_mem = global.player_mem[event.player_index]
		
		debug_print( "on_player_alt_selected_area" )
		
		if #event.entities == 0 then
			player_mem.clipboard = nil
			player.print({"copsel-gui-emptied"})
		else
			for i=1,#event.entities do
				local ent = event.entities[i]
				if ent.name ~= "straight-rail" and ent.name ~= "curved-rail" then
					player_mem.clipboard = ent
					if #event.entities > 1 then
						player.print({"copsel-gui-retaining",player_mem.clipboard.localised_name})
					end
					return
				end
			end
		end
	end
end

script.on_event(defines.events.on_player_alt_selected_area, on_player_alt_selected_area )

--------------------------------------------------------------------------------------
local function ask_wipe( player )
	local gui1 = player.gui.center.frm_copset
	
	if gui1 == nil then
		gui1 = player.gui.center.add({type = "frame", name = "frm_copset", caption = "Copy Settings Tool", direction = "vertical", style = "frame_copset_style"})
		gui1 = gui1.add({type = "flow", name = "flw_copset", direction = "vertical", style = "flow_copset_style"})		
	
		gui1.add({type = "label", name = "lbl_copset_wipes", caption = {"copset-gui-wipes"}, style = "label_copset_style"})	
		for n=1,#wipes do
			local name = wipes[n].name
			gui1.add({type = "button", name = "but_copset_wipe_"..name, caption = {"copset-gui-wipe-"..name}, style = "button_copset_style"})
		end
	end
end

--------------------------------------------------------------------------------------
function retrieve_in_bag(inv_player,inv)
	for n=1,#inv do
		if inv[n].valid_for_read then
			inv_player.insert(inv[n])
		end
	end
	return(not inv.is_empty())
end

--------------------------------------------------------------------------------------
function on_player_selected_area(event)
	-- paste settings of corresponding entities
	
	if event.item == "copy-settings-tool" then
		local player = game.players[event.player_index]
		local player_mem = global.player_mem[event.player_index]
 		local clipboard = player_mem.clipboard
		local source = nil
		local inv_player = player.get_inventory(defines.inventory.player_main)

		debug_print( "on_player_selected_area" )

		if clipboard and clipboard.valid then
			source = clipboard
		elseif player.entity_copy_source and player.entity_copy_source.valid then
			source = player.entity_copy_source
		end
		
		if source then
			-- multiple paste
			for i=1,#event.entities do
				local ent = event.entities[i]
				debug_print("paste ",ent.name)
				if ent ~= source then
					local results = ent.copy_settings(source)
					if inv_player then
						local s = ""
						for name, count in pairs(results) do
							inv_player.insert({name=name,count=count})
							s = s .. "+" .. name .. "(" .. count .. ") "
						end
						if s ~= "" then
							player.surface.create_entity({name = "flying-text", position = player.position, text = s, color = colors.white})
						end
					end
				end
			end
		else
			-- multiple wipe
			player_mem.targets = event.entities
			
			ask_wipe(player)
		end
	end
end

script.on_event(defines.events.on_player_selected_area, on_player_selected_area )

--------------------------------------------------------------------------------------
local function wipe_settings( player, type )
	local player_mem = global.player_mem[player.index]
	local inv_player = player.get_inventory(defines.inventory.player_main)
	local retrieved = false
	
	debug_print( "wipe ", wipes[type].name )
	
	for _, ent in pairs(player_mem.targets) do
		-- wipe slots of requesters
		if (type == wipes.all or type == wipes.requests) and ent.request_slot_count > 0 then
			for n=1,ent.request_slot_count do
				ent.clear_request_slot(n)
			end
		end
		
		-- wipe filters of inserter that have some
		if (type == wipes.all or type == wipes.filters) and ent.type == "inserter" and ent.filter_slot_count > 0 then
			for n=1,ent.filter_slot_count do
				ent.set_filter(n,nil)
			end
		end
		
		-- reset hashbars of chests
		if (type == wipes.all or type == wipes.hashbars) and (ent.type == "container" or ent.type == "logistic-container") then
			local inv = ent.get_inventory(defines.inventory.chest)
			if inv.hasbar() then
				inv.setbar()
			end
		end
		
		-- reset recipe of machines
		if (type == wipes.all or type == wipes.recipes) and ent.type == "assembling-machine" then
			retrieved = retrieve_in_bag(inv_player,ent.get_inventory(defines.inventory.assembling_machine_input)) or retrieved
			retrieved = retrieve_in_bag(inv_player,ent.get_inventory(defines.inventory.assembling_machine_output)) or retrieved
			ent.recipe = nil
		end
	end
	
	if retrieved then
		player.surface.create_entity({name = "flying-text", position = player.position, text = "+objects", color = colors.white})
	end		
end

--------------------------------------------------------------------------------------
local function on_gui_click(event)
	local player = game.players[event.player_index]
	local event_name = event.element.name
	
	-- debug_print( "on_gui_click ", player.name, " ", event_name )

	for n=1,#wipes do
		local name = wipes[n].name
		if event_name == "but_copset_wipe_"..name then
			if n < wipes.cancel then
				wipe_settings( player, n )
			end
			if player.gui.center.frm_copset then
				player.gui.center.frm_copset.destroy()
			end
		end
	end
end

script.on_event(defines.events.on_gui_click, on_gui_click)

--------------------------------------------------------------------------------------
function on_hotkey_main(event)
	local player = game.players[event.player_index]
	if player.cursor_stack and player.cursor_stack.valid_for_read and player.cursor_stack.name == "copy-settings-tool" then
		player.clean_cursor()
		-- player.cursor_stack.clear()
		-- player.insert({name = "copy-settings-tool", count = 1})
	elseif player.get_item_count("copy-settings-tool") > 0 then
		player.cursor_stack.set_stack({name = "copy-settings-tool", count = 1})
		player.remove_item({name = "copy-settings-tool", count = 1})
	end
end

script.on_event("copset_hotkey", on_hotkey_main)

--------------------------------------------------------------------------------------
local function on_pre_entity_settings_pasted( event )
	local ent_source = event.source
	local ent_destination = event.destination
	
	debug_print( "on_pre_entity_settings_pasted ", ent_source.name, " -> ", ent_destination.name)
end

-- script.on_event(defines.events.on_pre_entity_settings_pasted, on_pre_entity_settings_pasted )

